@@ -20,11 +20,13 @@
 #
 #
 
+import casadi as cs
 from casadi import *
 from collections import defaultdict, OrderedDict
+from contextlib import contextmanager
 
 def get_ranges_dict(list_expr):
-    ret = {}
+    ret = HashDict()
     offset = 0
     for e in list_expr:
         next_offset = offset+e.nnz()
@@ -127,6 +129,12 @@ def single_stacktrace(m):
     m["stacktrace"] = m["stacktrace"][0]
     return m
 
+def reshape_number(target, value):
+    if not isinstance(value,cs.MX):
+        value = cs.DM(value)
+    if value.is_scalar():
+        value = cs.DM.ones(target.shape)*value
+    return value
 
 
 def DM2numpy(dm, expr_shape, tdim=None):
@@ -277,4 +285,169 @@ def for_all_primitives(expr, rhs, callback, msg, rhs_type=MX):
             offset += p.nnz()
     else:
         raise Exception(msg)
+        
+
+class Node:
+  def __init__(self,val):
+    self.val = val
+    self.nodes = []
+
+class AutoBrancher:
+  OPEN = 0
+  DONE = 1
+  def __init__(self):
+    self.root = Node(AutoBrancher.OPEN)
+    self.trace = [self.root]
+
+  @property
+  def current(self):
+    return self.trace[-1]
+
+  def branch(self, alternatives = [True, False]):
+    alternatives = list(alternatives)
+    nodes = self.current.nodes
+    if len(nodes)==0:
+      nodes += [None]*len(alternatives)
+    for i,n in enumerate(nodes):
+      if n is None:
+        nodes[i] = Node(AutoBrancher.OPEN)
+        self.trace.append(nodes[i])
+        self.this_branch.append(alternatives[i])
+        return alternatives[i]
+      else:
+        if n.val == AutoBrancher.OPEN:
+          self.trace.append(nodes[i])
+          self.this_branch.append(alternatives[i])
+          return alternatives[i]
+
+  def __iter__(self):
+    cnt = 0
+    
+    while self.root.val==AutoBrancher.OPEN:
+      self.this_branch = []
+      cnt+=1
+      yield self
+      # Indicate that current leaf is done
+      self.current.val = AutoBrancher.DONE
+      # Close leaves when subleaves are done
+      for n in reversed(self.trace[:-1]):
+        finished = True
+        for e in n.nodes:
+          finished = finished and e and e.val==AutoBrancher.DONE
+        if finished:
+          n.val = AutoBrancher.DONE
+      # Reset trace
+      self.trace = [self.root]
+      print("Evaluated branch",self.this_branch)
+    print("Evaluated",cnt,"branches")
+
+def vvcat(arg):
+    if len(arg)==0:
+        return cs.MX(0,1)
+    else:
+        return cs.vvcat(arg)
+
+def vcat(arg):
+    if len(arg)==0:
+        return cs.MX(0,1)
+    else:
+        return cs.vcat(arg)
+
+def prepare_build_dir(build_dir_abs):
+    import os
+    import shutil
+    os.makedirs(build_dir_abs,exist_ok=True)
+    # Clean directory (but do not delete it,
+    # since this confuses open shells in Linux (e.g. bash, Matlab)
+    for filename in os.listdir(build_dir_abs):
+      file_path = os.path.join(build_dir_abs, filename)
+      try:
+        if os.path.isfile(file_path) or os.path.islink(file_path):
+          os.unlink(file_path)
+        elif os.path.isdir(file_path):
+          shutil.rmtree(file_path)
+      except:
+        pass
+    
+
+
+class ConstraintInspector:
+    def __init__(self, method, stage):
+        self.opti = Opti()
+
+        self.X = self.opti.variable(*stage.x.shape)
+        self.U = self.opti.variable(*stage.u.shape)
+        self.V = self.opti.variable(*stage.v.shape)
+        self.P = self.opti.parameter(*stage.p.shape)
+        self.t = self.opti.variable()
+        self.T = self.opti.variable()
+
+        offsets = list(stage._offsets.keys())
+
+        self.offsets = []
+        for e in offsets:
+            self.offsets.append(self.opti.variable(*e.shape))
+
+        self.raw = [stage.x,stage.u,stage.p,stage.t, method.v]+offsets
+        self.optivar = [self.X, self.U, self.P, self.t, self.V]+self.offsets
+
+        if method.free_time:
+            self.raw += [stage.T]
+            self.optivar += [self.T]
+    
+    def finalize(self):
+        self.opti_advanced = self.opti.advanced
+
+    def canon(self,expr):
+        c = substitute([expr],self.raw,self.optivar)[0]
+        mc = self.opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
+        return substitute([mc.lb,mc.canon,mc.ub],self.optivar,self.raw), mc
+
+
+def linear_coeffs(expr, *args):
+    """ Multi-argument extesion to CasADi linear_coeff"""
+    J,c = linear_coeff(expr, vcat(args))
+    cs = np.cumsum([0]+[e.numel() for e in args])
+    return tuple([J[:,cs[i]:cs[i+1]] for i in range(len(args))])+(c,)
+
+ca_classes = [cs.SX,cs.MX,cs.Function,cs.Sparsity,cs.DM,cs.Opti]
+
+@contextmanager
+def rockit_pickle_context():
+    string_serializer = cs.StringSerializer()
+    string_serializer.pack("preamble")
+    string_serializer.encode()
+    def __getstate__(self):
+        if isinstance(self,cs.Opti):
+            raise Exception("Opti cannot be serialized yet. Consider removing the transcribed problem.")
+        string_serializer.pack(self)
+        enc = string_serializer.encode()
+        return {"s":enc}
+
+    for c in ca_classes:
+        setattr(c, '__getstate__', __getstate__)
+        
+    yield
+
+    for c in ca_classes:
+        delattr(c, '__getstate__')
+
+@contextmanager
+def rockit_unpickle_context():
+    string_serializer = cs.StringSerializer()
+    string_serializer.pack("preamble")
+    string_deserializer = cs.StringDeserializer(string_serializer.encode())
+    string_deserializer.unpack()
+    def __setstate__(self,state):
+        string_deserializer.decode(state["s"])
+        s = string_deserializer.unpack()
+        self.this = s.this
+
+    for c in ca_classes:
+        setattr(c, '__setstate__', __setstate__)
+        
+    yield
+
+    for c in ca_classes:
+        delattr(c, '__setstate__')
         
\ No newline at end of file
