@@ -48,15 +48,17 @@ class FixedGrid(Grid):
         self.localize_t0 = localize_t0
         self.localize_T = localize_T
 
-    def bounds_T(self, opti, T_local, t0_local, k, T, N):
+    def bounds_T(self, T_local, t0_local, k, T, N):
         if self.localize_T:
             Tk = T_local[k]
             if k>=0 and k+1<N:
-                self.constrain_T(opti, T_local[k], T_local[k+1], N)
+                r = self.constrain_T(T_local[k], T_local[k+1], N)
+                if r is not None:
+                    yield r
         else:
             Tk = T*(self.normalized(N)[k+1]-self.normalized(N)[k])
         if self.localize_t0 and k>=0:
-            opti.subject_to(t0_local[k]+Tk==t0_local[k+1])
+            yield (t0_local[k]+Tk==t0_local[k+1],{})
 
     def get_t0_local(self, opti, k, t0, N):
         if self.localize_t0:
@@ -94,12 +96,13 @@ class FreeGrid(FixedGrid):
         FixedGrid.__init__(self, **kwargs)
         self.localize_T = True
 
-    def constrain_T(self,opti,T,Tnext,N):
+    def constrain_T(self,T,Tnext,N):
         pass
 
-    def bounds_T(self, opti, T_local, t0_local, k, T, N):
-        opti.subject_to(self.min <= (T_local[k] <= self.max))
-        FixedGrid.bounds_T(self, opti, T_local, t0_local, k, T, N)
+    def bounds_T(self, T_local, t0_local, k, T, N):
+        yield (self.min <= (T_local[k] <= self.max),{})
+        for e in FixedGrid.bounds_T(self, T_local, t0_local, k, T, N):
+            yield e
 
     def bounds_finalize(self, opti, control_grid, t0_local, tf, N):
         opti.subject_to(control_grid[-1]==tf)
@@ -124,19 +127,23 @@ class UniformGrid(FixedGrid):
     def __init__(self, **kwargs):
         FixedGrid.__init__(self, **kwargs)
 
-    def constrain_T(self,opti,T,Tnext,N):
-        opti.subject_to(T==Tnext)
+    def constrain_T(self,T,Tnext,N):
+        return (T==Tnext,{})
 
     def scale_first(self, N):
         return 1.0/N
 
-    def bounds_T(self, opti, T_local, t0_local, k, T, N):
+    def bounds_T(self, T_local, t0_local, k, T, N):
         if k==0:
             if self.localize_T:
-                opti.subject_to(self.min <= (T_local[0] <= self.max))
+                yield (self.min <= (T_local[0] <= self.max), {})
             else:
-                opti.subject_to(self.min <= (T/N <= self.max))
-        FixedGrid.bounds_T(self, opti, T_local, t0_local, k, T, N)
+                if self.min==0 and self.max==inf:
+                    pass
+                else:
+                    yield (self.min <= (T/N <= self.max), {})
+        for e in FixedGrid.bounds_T(self, T_local, t0_local, k, T, N):
+            yield e
 
     def normalized(self, N):
         return list(np.linspace(0.0, 1.0, N+1))
@@ -184,23 +191,24 @@ class GeometricGrid(FixedGrid):
             return self._growth_factor**(1.0/(N-1))
         return self._growth_factor
 
-    def constrain_T(self,opti,T,Tnext,N):
-        opti.subject_to(T*self.growth_factor(N)==Tnext)
+    def constrain_T(self,T,Tnext,N):
+        return (T*self.growth_factor(N)==Tnext,{})
 
     def scale_first(self, N):
         return self.normalized(N)[1]
 
-    def bounds_T(self, opti, T_local, t0_local, k, T, N):
+    def bounds_T(self, T_local, t0_local, k, T, N):
         if self.localize_T:
             if k==0 or k==-1:
-                opti.subject_to(self.min <= (T_local[k] <= self.max))
+                yield (self.min <= (T_local[k] <= self.max), {})
         else:
             n = self.normalized(N)
             if k==0:
-                opti.subject_to(self.min <= (T*n[1] <= self.max))
+                yield (self.min <= (T*n[1] <= self.max),{})
             if k==-1:
-                opti.subject_to(self.min <= (T*(n[-1]-n[-2]) <= self.max))
-        FixedGrid.bounds_T(self, opti, T_local, t0_local, k, T, N)
+                yield (self.min <= (T*(n[-1]-n[-2]) <= self.max),{})
+        for e in FixedGrid.bounds_T(self, T_local, t0_local, k, T, N):
+            yield e
 
     def normalized(self, N):
         g = self.growth_factor(N)
@@ -224,6 +232,9 @@ class SamplingMethod(DirectMethod):
             Name for CasADi Function
         options : dict, optional
             Options for CasADi Function
+        intg : `str`
+            Rockit-specific methods: 'rk', 'expl_euler' - these allow to make use of signal sampling with 'refine'
+            Others are passed on as-is to CasADi
         """
         DirectMethod.__init__(self, **kwargs)
         self.N = N
@@ -231,7 +242,9 @@ class SamplingMethod(DirectMethod):
         self.intg = intg
         self.intg_options = {} if intg_options is None else intg_options
         self.time_grid = grid
+        self.clean()
 
+    def clean(self):
         self.X = []  # List that will hold N+1 decision variables for state vector
         self.U = []  # List that will hold N decision variables for control vector
         self.Z = []  # Algebraic vars
@@ -242,8 +255,10 @@ class SamplingMethod(DirectMethod):
         self.P = []
         self.V = None
         self.V_control = []
+        self.V_control_plus = []
         self.V_states = []
         self.P_control = []
+        self.P_control_plus = []
 
         self.poly_coeff = []  # Optional list to save the coefficients for a polynomial
         self.poly_coeff_z = []  # Optional list to save the coefficients for a polynomial
@@ -283,10 +298,11 @@ class SamplingMethod(DirectMethod):
             intg = getattr(self, "intg_" + self.intg)(stage._ode(), X0, U, P, Z)
         else:
             intg = self.intg_builtin(stage._ode(), X0, U, P, Z)
-        assert not intg.has_free()
+        if intg.has_free():
+            raise Exception("Free variables found: %s" % str(intg.get_free()))
     
         for j in range(self.M):
-            intg_res = intg(x0=X[-1], u=U, t0=t0_local, DT=DT, p=P)
+            intg_res = intg(x0=X[-1], u=U, t0=t0_local, DT=DT, DT_control=T, p=P)
             X.append(intg_res["xf"])
             Zs.append(intg_res["zf"])
             quad = quad + intg_res["qf"]
@@ -302,6 +318,7 @@ class SamplingMethod(DirectMethod):
     def intg_rk(self, f, X, U, P, Z):
         assert Z.is_empty()
         DT = MX.sym("DT")
+        DT_control = MX.sym("DT_control")
         t0 = MX.sym("t0")
         # A single Runge-Kutta 4 step
         k1 = f(x=X, u=U, p=P, t=t0)
@@ -314,23 +331,36 @@ class SamplingMethod(DirectMethod):
         f2 = 4/DT**2*(k3["ode"]-k2["ode"])/6
         f3 = 4*(k4["ode"]-2*k3["ode"]+k1["ode"])/DT**3/24
         poly_coeff = hcat([X, f0, f1, f2, f3])
-        return Function('F', [X, U, t0, DT, P], [X + DT / 6 * (k1["ode"] + 2 * k2["ode"] + 2 * k3["ode"] + k4["ode"]), poly_coeff, DT / 6 * (k1["quad"] + 2 * k2["quad"] + 2 * k3["quad"] + k4["quad"]), MX(0, 1), MX()], ['x0', 'u', 't0', 'DT', 'p'], ['xf', 'poly_coeff', 'qf', 'zf', 'poly_coeff_z'])
+        return Function('F', [X, U, t0, DT, DT_control, P], [X + DT / 6 * (k1["ode"] + 2 * k2["ode"] + 2 * k3["ode"] + k4["ode"]), poly_coeff, DT / 6 * (k1["quad"] + 2 * k2["quad"] + 2 * k3["quad"] + k4["quad"]), MX(0, 1), MX()], ['x0', 'u', 't0', 'DT', 'DT_control', 'p'], ['xf', 'poly_coeff', 'qf', 'zf', 'poly_coeff_z'])
+
+    def intg_expl_euler(self, f, X, U, P, Z):
+        assert Z.is_empty()
+        DT = MX.sym("DT")
+        DT_control = MX.sym("DT_control")
+        t0 = MX.sym("t0")
+        k = f(x=X, u=U, p=P, t=t0)
+        poly_coeff = hcat([X, k["ode"]])
+        return Function('F', [X, U, t0, DT, DT_control, P], [X + DT * k["ode"], poly_coeff, DT * k["quad"], MX(0, 1), MX()], ['x0', 'u', 't0', 'DT', 'DT_control', 'p'], ['xf', 'poly_coeff', 'qf', 'zf', 'poly_coeff_z'])
 
     def intg_builtin(self, f, X, U, P, Z):
         # A single CVODES step
         DT = MX.sym("DT")
+        DT_control = MX.sym("DT_control")
         t = MX.sym("t")
         t0 = MX.sym("t0")
         res = f(x=X, u=U, p=P, t=t0+t*DT, z=Z)
-        data = {'x': X, 'p': vertcat(U, DT, P, t0), 'z': Z, 't': t, 'ode': DT * res["ode"], 'quad': DT * res["quad"], 'alg': res["alg"]}
+        data = {'x': X, 'p': vertcat(U, DT, DT_control, P, t0), 'z': Z, 't': t, 'ode': DT * res["ode"], 'quad': DT * res["quad"], 'alg': res["alg"]}
         options = dict(self.intg_options)
         if self.intg in ["collocation"]:
             # In rockit, M replaces number_of_finite_elements on a higher level
             if "number_of_finite_elements" not in options:
                 options["number_of_finite_elements"] = 1
         I = integrator('intg_'+self.intg, self.intg, data, options)
-        res = I.call({'x0': X, 'p': vertcat(U, DT, P, t0)})
-        return Function('F', [X, U, t0, DT, P], [res["xf"], MX(), res["qf"], res["zf"], MX()], ['x0', 'u', 't0', 'DT', 'p'], ['xf', 'poly_coeff','qf','zf','poly_coeff_z'])
+        res = I.call({'x0': X, 'p': vertcat(U, DT, DT_control, P, t0)})
+        return Function('F', [X, U, t0, DT, DT_control, P], [res["xf"], MX(), res["qf"], res["zf"], MX()], ['x0', 'u', 't0', 'DT', 'DT_control', 'p'], ['xf', 'poly_coeff','qf','zf','poly_coeff_z'])
+
+    def untranscribe_placeholders(self, phase, stage):
+        pass
 
     def transcribe_placeholders(self, phase, stage, placeholders):
         """
@@ -338,6 +368,12 @@ class SamplingMethod(DirectMethod):
         """
         return stage._transcribe_placeholders(phase, self, placeholders)
 
+    def transcribe_start(self, stage, opti):
+        return
+        
+    def untranscribe(self, stage, phase=1,**kwargs):
+        self.clean()
+
     def transcribe(self, stage, phase=1,**kwargs):
         """
         Transcription is the process of going from a continuous-time OCP to an NLP
@@ -346,8 +382,12 @@ class SamplingMethod(DirectMethod):
         if phase>1: return
         opti = stage.master._method.opti
         DM.set_precision(14)
-        self.add_variables(stage, opti)
+
+        self.transcribe_start(stage, opti)
+        # Parameters needed before variables because of self.T = self.eval(stage, stage._T)
         self.add_parameter(stage, opti)
+        self.set_parameter(stage, opti)
+        self.add_variables(stage, opti)
 
         self.integrator_grid = []
         for k in range(self.N):
@@ -380,19 +420,24 @@ class SamplingMethod(DirectMethod):
 
 
     def add_constraints_before(self, stage, opti):
-        for c, meta, _ in stage._constraints["point"]:
+        for c, meta, args in stage._constraints["point"]:
             e = self.eval(stage, c)
             if 'r_at_tf' not in [a.name() for a in symvar(e)]:
-                opti.subject_to(e, meta=meta)
+                opti.subject_to(e, args["scale"], meta=meta)
 
     def add_constraints_after(self, stage, opti):
-        for c, meta, _ in stage._constraints["point"]:
+        for c, meta, args in stage._constraints["point"]:
             e = self.eval(stage, c)
             if 'r_at_tf' in [a.name() for a in symvar(e)]:
-                opti.subject_to(e, meta=meta)
+                opti.subject_to(e, args["scale"], meta=meta)
 
     def add_inf_constraints(self, stage, opti, c, k, l, meta):
+        # Query the discretization method used for polynomial coefficients
+        #   interpretation: state ~= coeff * [t^0;t^1;t^2;...]
+        #                    t is physical time, but starting at 0 at the beginning of the interval
         coeff = stage._method.poly_coeff[k * self.M + l]
+
+        # Represent polynomial as a BSpline object (https://gitlab.kuleuven.be/meco-software/rockit/-/blob/v0.1.28/rockit/splines/spline.py#L392)
         degree = coeff.shape[1]-1
         basis = BSplineBasis([0]*(degree+1)+[1]*(degree+1),degree)
         tscale = self.T / self.N / self.M
@@ -400,25 +445,34 @@ class SamplingMethod(DirectMethod):
         coeff = coeff * repmat(tpower.T,stage.nx,1)
         # TODO: bernstein transformation as function of degree
         Poly_to_Bernstein_matrix_4 = DM([[1,0,0,0,0],[1,1.0/4, 0, 0, 0],[1, 1.0/2, 1.0/6, 0, 0],[1, 3.0/4, 1.0/2, 1.0/4, 0],[1, 1, 1, 1, 1]])
+        # Use a direct way to obtain Bernstein coefficients from polynomial coefficients
         state_coeff = mtimes(Poly_to_Bernstein_matrix_4,coeff.T)
         
+        # Replace symbols for states by BSpline object derivatives
+        subst_from = list(stage.states)
+
         statesize = [0] + [elem.nnz() for elem in stage.states]
         statessizecum = np.cumsum(statesize)
-
-        subst_from = list(stage.states)
         state_coeff_split = horzsplit(state_coeff,statessizecum)
         subst_to = [BSpline(basis,coeff) for coeff in state_coeff_split]
 
 
         lookup = dict(zip(subst_from, subst_to))
 
+        # Replace symbols for state derivatives by BSpline object derivatives
         subst_from += stage._inf_der.keys()
         dt = (self.control_grid[k + 1] - self.control_grid[k])/self.M
         subst_to += [lookup[e].derivative()*(1/dt) for e in stage._inf_der.values()]
 
         subst_from += stage._inf_inert.keys()
         subst_to += stage._inf_inert.values()
+
+        # Here, the actual replacement takes place
         c_spline = reinterpret_expr(c, subst_from, subst_to)
+
+        # The use of '>=' or '<=' on BSpline objects automatically results in
+        # those same operations being relayed onto BSpline coefficients
+        # see https://gitlab.kuleuven.be/meco-software/rockit/-/blob/v0.1.28/rockit/splines/spline.py#L520-521
         try:
             opti.subject_to(self.eval_at_control(stage, c_spline, k), meta=meta)
         except IndexError:
@@ -438,6 +492,13 @@ class SamplingMethod(DirectMethod):
             r = r + self.eval_at_control(stage, expr, k)
         return r
 
+    def fill_placeholders_sum_control_plus(self, phase, stage, expr, *args):
+        if phase==1: return
+        r = 0
+        for k in list(range(self.N))+[-1]:
+            r = r + self.eval_at_control(stage, expr, k)
+        return r
+
     def placeholders_next(self, stage, expr, *args):
         self.eval_at_control(stage, expr, k+1)
 
@@ -465,17 +526,22 @@ class SamplingMethod(DirectMethod):
     def add_variables_V_control(self, stage, opti, k):
         if k==0:
             self.V_control = [[] for v in stage.variables['control']]
+            self.V_control_plus = [[] for v in stage.variables['control+']]
             for i, v in enumerate(stage.variables['states']):
-                self.V_states.append([opti.variable(v.shape[0], v.shape[1])])
+                self.V_states.append([opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v])])
         for i, v in enumerate(stage.variables['control']):
-            self.V_control[i].append(opti.variable(v.shape[0], v.shape[1]))
+            self.V_control[i].append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v]))
+        for i, v in enumerate(stage.variables['control+']):
+            self.V_control_plus[i].append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v]))
         for i, v in enumerate(stage.variables['states']):
-            self.V_states[i].append(opti.variable(v.shape[0], v.shape[1]))
+            self.V_states[i].append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v]))
 
         self.t0_local[k] = self.time_grid.get_t0_local(opti, k, self.t0, self.N)
         self.T_local[k] = self.time_grid.get_T_local(opti, k, self.T, self.N)
 
     def add_variables_V_control_finalize(self, stage, opti):
+        for i, v in enumerate(stage.variables['control+']):
+            self.V_control_plus[i].append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v]))
         if self.time_grid.localize_t0:
             self.t0_local[self.N] = opti.variable()
             self.control_grid = hcat(self.t0_local)
@@ -492,7 +558,10 @@ class SamplingMethod(DirectMethod):
         self.time_grid.bounds_finalize(opti, self.control_grid, self.t0_local, self.t0+self.T, self.N)
 
     def add_coupling_constraints(self, stage, opti, k):
-        self.time_grid.bounds_T(opti, self.T_local, self.t0_local, k, self.T, self.N)
+        advanced = opti.advanced
+        for c,kwargs in self.time_grid.bounds_T(self.T_local, self.t0_local, k, self.T, self.N):
+            if not advanced.is_parametric(c):
+                opti.subject_to(c,**kwargs)
 
     def get_p_control_at(self, stage, k=-1):
         return veccat(*[p[k] for p in self.P_control])
@@ -500,11 +569,17 @@ class SamplingMethod(DirectMethod):
     def get_v_control_at(self, stage, k=-1):
         return veccat(*[v[k] for v in self.V_control])
 
+    def get_p_control_plus_at(self, stage, k=-1):
+        return veccat(*[p[k] for p in self.P_control_plus])
+
+    def get_v_control_plus_at(self, stage, k=-1):
+        return veccat(*[v[k] for v in self.V_control_plus])
+
     def get_v_states_at(self, stage, k=-1):
         return veccat(*[v[k] for v in self.V_states])
 
     def get_p_sys(self, stage, k):
-        return vertcat(vvcat(self.P), self.get_p_control_at(stage, k), self.V, self.get_v_control_at(stage, k))
+        return vertcat(vvcat(self.P), self.get_p_control_at(stage, k), self.get_p_control_plus_at(stage, k), self.V, self.get_v_control_at(stage, k), self.get_v_control_plus_at(stage, k))
 
     def eval(self, stage, expr):
         return stage.master._method.eval_top(stage.master, stage._expr_apply(expr, p=veccat(*self.P), v=self.V, t0=stage.t0, T=stage.T))
@@ -533,60 +608,53 @@ class SamplingMethod(DirectMethod):
             subst_to.append(self._eval_at_control(stage, vvcat(offsets[offset]), k+offset))
             #print(expr, subst_from, subst_to)
 
-        expr = stage._expr_apply(expr, sub=(subst_from, subst_to), t0=self.t0, T=self.T, x=self.X[k], z=self.Z[k] if self.Z else nan, xq=self.q if k==-1 else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k),  v_states=self.get_v_states_at(stage, k), t=self.control_grid[k])
+
+        DT_control = self.get_DT_control_at(k)
+        DT = self.get_DT_at(k, self.M-1 if k==-1 else 0)
+
+        expr = stage._expr_apply(expr, sub=(subst_from, subst_to), t0=self.t0, T=self.T, x=self.X[k], z=self.Z[k] if self.Z else nan, xq=self.q if k==-1 else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), p_control_plus=self.get_p_control_plus_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k),  v_control_plus=self.get_v_control_plus_at(stage, k), v_states=self.get_v_states_at(stage, k), t=self.control_grid[k], DT=DT, DT_control=DT_control)
         expr = stage.master._method.eval_top(stage.master, expr)
         #print("expr",expr)
         return expr
+    
+    def get_DT_control_at(self, k):
+        if k==-1 or k==self.N:
+            return self.control_grid[-1]-self.control_grid[-2]
+        return self.control_grid[k+1]-self.control_grid[k]
+
+    def get_DT_at(self, k, i):
+        integrator_grid = self.integrator_grid[k]
+        if i<integrator_grid.numel()-1:
+            return integrator_grid[i+1]-integrator_grid[i]
+        else:
+            return self.integrator_grid[k+1][0]-self.integrator_grid[k][i]
 
     def _eval_at_control(self, stage, expr, k):
         x = self.X[k]
         z = self.Z[k] if self.Z else nan
         u = self.U[-1] if k==len(self.U) else self.U[k] # Would be fixed if we remove the (k=-1 case)
-        p_control = self.get_p_control_at(stage, k)
-        v_control = self.get_v_control_at(stage, k)
+        p_control = self.get_p_control_at(stage, k) if k!=len(self.U) else self.get_p_control_at(stage, k-1)
+        v_control = self.get_v_control_at(stage, k) if k!=len(self.U) else self.get_v_control_at(stage, k-1)
+        p_control_plus = self.get_p_control_plus_at(stage, k)
+        v_control_plus = self.get_v_control_plus_at(stage, k)
         v_states = self.get_v_states_at(stage, k)
         t = self.control_grid[k]
-        return stage._expr_apply(expr, t0=self.t0, T=self.T, x=x, z=z, xq=self.q if k==-1 else nan, u=u, p_control=p_control, v=self.V, p=veccat(*self.P), v_control=v_control,  v_states=v_states, t=t)
-        kwargs = dict(t0=self.t0, T=self.T, v=self.V, p=veccat(*self.P))
-        try:
-            kwargs["x"] = self.X[k]
-        except IndexError:
-            pass
-        try:
-            kwargs["z"] = self.Z[k] if self.Z else nan
-        except IndexError:
-            pass
-        try:
-            kwargs["xq"] = self.q if k==-1 else nan
-        except IndexError:
-            pass
-        try:
-            kwargs["u"] = self.U[k]
-        except IndexError:
-            pass
-        try:
-            kwargs["p_control"] = self.get_p_control_at(stage, k)
-        except IndexError:
-            pass
-        try:
-            kwargs["v_control"] = self.get_v_control_at(stage, k)
-        except IndexError:
-            pass
-        try:
-            kwargs["v_states"] = self.get_v_states_at(stage, k)
-        except IndexError:
-            pass
-        try:
-            kwargs["t"] = self.control_grid[k]
-        except IndexError:
-            pass
-        return stage._expr_apply(expr, **kwargs)
+        DT_control = self.get_DT_control_at(k)
+        if k==-1 or k==len(self.integrator_grid):
+            DT = self.get_DT_at(len(self.integrator_grid)-1, self.M-1)
+        else:
+            DT = self.get_DT_at(k, 0)
+        return stage._expr_apply(expr, t0=self.t0, T=self.T, x=x, z=z, xq=self.q if k==-1 else nan, u=u, p_control=p_control, p_control_plus=p_control_plus, v=self.V, p=veccat(*self.P), v_control=v_control, v_control_plus=v_control_plus, v_states=v_states, t=t, DT=DT, DT_control=DT_control)
 
     def eval_at_integrator(self, stage, expr, k, i):
-        return stage.master._method.eval_top(stage.master, stage._expr_apply(expr, t0=self.t0, T=self.T, x=self.xk[k*self.M + i], z=self.zk[k*self.M + i] if self.zk else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k),  v_states=self.get_v_states_at(stage, k), t=self.integrator_grid[k][i]))
+        DT_control = self.get_DT_control_at(k)
+        DT = self.get_DT_at(k, i)
+        return stage.master._method.eval_top(stage.master, stage._expr_apply(expr, t0=self.t0, T=self.T, x=self.xk[k*self.M + i], z=self.zk[k*self.M + i] if self.zk else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), p_control_plus=self.get_p_control_plus_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k), v_control_plus=self.get_v_control_plus_at(stage, k), v_states=self.get_v_states_at(stage, k), t=self.integrator_grid[k][i], DT=DT, DT_control=DT_control))
 
     def eval_at_integrator_root(self, stage, expr, k, i, j):
-        return stage.master._method.eval_top(stage.master, stage._expr_apply(expr, t0=self.t0, T=self.T, x=self.xr[k][i][:,j], z=self.zr[k][i][:,j] if self.zk else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k), t=self.tr[k][i][j]))
+        DT_control = self.get_DT_control_at(k)
+        DT = self.get_DT_at(k, i)
+        return stage.master._method.eval_top(stage.master, stage._expr_apply(expr, t0=self.t0, T=self.T, x=self.xr[k][i][:,j], z=self.zr[k][i][:,j] if self.zk else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), p_control_plus=self.get_p_control_plus_at(stage, k),v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k), v_control_plus=self.get_v_control_plus_at(stage, k),t=self.tr[k][i][j], DT=DT, DT_control=DT_control))
 
     def set_initial(self, stage, master, initial):
         opti = master.opti if hasattr(master, 'opti') else master
@@ -624,6 +692,10 @@ class SamplingMethod(DirectMethod):
             if is_equal(parameter, p):
                 found = True
                 opti.set_value(hcat(self.P_control[i]), value)
+        for i, p in enumerate(stage.parameters['control+']):
+            if is_equal(parameter, p):
+                found = True
+                opti.set_value(hcat(self.P_control_plus[i]), value)
         assert found, "You attempted to set the value of a non-parameter."
 
     def add_parameter(self, stage, opti):
@@ -631,9 +703,13 @@ class SamplingMethod(DirectMethod):
             self.P.append(opti.parameter(p.shape[0], p.shape[1]))
         for p in stage.parameters['control']:
             self.P_control.append([opti.parameter(p.shape[0], p.shape[1]) for i in range(self.N)])
+        for p in stage.parameters['control+']:
+            self.P_control_plus.append([opti.parameter(p.shape[0], p.shape[1]) for i in range(self.N+1)])
 
     def set_parameter(self, stage, opti):
         for i, p in enumerate(stage.parameters['']):
             opti.set_value(self.P[i], stage._param_value(p))
         for i, p in enumerate(stage.parameters['control']):
-            opti.set_value(hcat(self.P_control[i]), stage._param_value(p))
\ No newline at end of file
+            opti.set_value(hcat(self.P_control[i]), stage._param_value(p))
+        for i, p in enumerate(stage.parameters['control+']):
+            opti.set_value(hcat(self.P_control_plus[i]), stage._param_value(p))
\ No newline at end of file
