@@ -19,10 +19,10 @@
 #     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 #
 #
-
 from casadi import MX, substitute, Function, vcat, depends_on, vertcat, jacobian, veccat, jtimes, hcat,\
-                   linspace, DM, constpow, mtimes, vvcat, low, floor, hcat, horzcat, DM, is_equal, \
+                   linspace, DM, constpow, mtimes, low, floor, hcat, horzcat, DM, is_equal, \
                    Sparsity
+from rockit.grouping_techniques import GroupingTechnique
 from .freetime import FreeTime
 from .direct_method import DirectMethod
 from .multiple_shooting import MultipleShooting
@@ -31,6 +31,7 @@ from collections import defaultdict
 from .casadi_helpers import DM2numpy, get_meta, merge_meta, HashDict, HashDefaultDict, HashOrderedDict, HashList, for_all_primitives
 from contextlib import contextmanager
 from collections import OrderedDict
+from .casadi_helpers import vvcat
 
 import numpy as np
 from numpy import nan
@@ -84,17 +85,21 @@ class Stage:
         self.parent = parent
 
         self._meta = HashDict()
+        self._scale = HashDict()
         self._var_original = None
         self._var_augmented = None
 
         self._param_vals = HashDict()
         self._state_der = HashDict()
+        self._scale_der = HashDict()
         self._state_next = HashDict()
         self._alg = []
         self._constraints = defaultdict(list)
         self._objective = 0
         self._initial = HashOrderedDict()
 
+        self._catalog = HashDict()
+
         self._placeholders = HashDict()
         self._offsets = HashDict()
         self._inf_inert = HashOrderedDict()
@@ -107,7 +112,9 @@ class Stage:
         self._public_T  = self._create_placeholder_expr(0, 'T')
         self._public_t0 = self._create_placeholder_expr(0, 't0')
         self._tf = self.T + self.t0
-    
+        self._public_DT = self._create_placeholder_expr(0, 'DT')
+        self._public_DT_control = self._create_placeholder_expr(0, 'DT_control')
+
     @property
     def master(self):
         return self._master
@@ -127,6 +134,14 @@ class Stage:
     @property
     def tf(self):
         return self._tf
+    
+    @property 
+    def DT(self):
+        return self._public_DT
+
+    @property 
+    def DT_control(self):
+        return self._public_DT_control
 
     def set_t0(self, t0):
         self._t0 = t0
@@ -167,7 +182,14 @@ class Stage:
         self._set_transcribed(False)
         return s
 
-    def state(self, n_rows=1, n_cols=1, quad=False,meta=None):
+    def _parse_scale(self, e, scale):
+        if DM(scale).is_scalar():
+            return DM.ones(e.sparsity())*scale
+        else:
+            assert scale.shape==e.shape
+            return scale
+
+    def state(self, n_rows=1, n_cols=1, quad=False, scale=1, meta=None):
         """Create a state.
         You must supply a derivative for the state with :obj:`~rockit.stage.Stage.set_der`
 
@@ -200,15 +222,26 @@ class Stage:
 
         name = "q"+str(len(self.qstates)+1) if quad else "x"+str(len(self.states)+1)
         x = MX.sym(name, n_rows, n_cols)
+        meta = merge_meta(meta, get_meta())
+        return self.register_state(x, meta=meta, quad=quad, scale=scale)
+        
+    def register_state(self, x, quad=False, scale=1, meta=None):
+        if isinstance(x, list):
+            for e in x:
+                self.register_state(e, quad=quad, scale=scale, meta=meta)
+            return
         self._meta[x] = merge_meta(meta, get_meta())
+        self._scale[x] = self._parse_scale(x, scale)
         if quad:
+            self._catalog[x] = {"type": 'qstates', "sparsity": x.sparsity()}
             self.qstates.append(x)
         else:
+            self._catalog[x] = {"type": 'states', "sparsity": x.sparsity()}
             self.states.append(x)
         self._set_transcribed(False)
         return x
 
-    def algebraic(self, n_rows=1, n_cols=1,meta=None):
+    def algebraic(self, n_rows=1, n_cols=1, scale=1, meta=None):
         """Create an algebraic variable
         You must supply an algebraic relation with:obj:`~rockit.stage.Stage.set_alg`
 
@@ -228,31 +261,146 @@ class Stage:
         """
         # Create a placeholder symbol with a dummy name (see #25)
         z = MX.sym("z", n_rows, n_cols)
+        meta = merge_meta(meta, get_meta())
+        return self.register_algebraic(z, scale=scale, meta=meta)
+
+    def register_algebraic(self, z, scale=1, meta=None):
+        if isinstance(z, list):
+            for e in z:
+                self.register_algebraic(e, scale=scale, meta=meta)
+            return
         self._meta[z] = merge_meta(meta, get_meta())
+        self._scale[z] = self._parse_scale(z, scale)
+        self._catalog[z] = {"type": 'algebraics', "sparsity": z.sparsity()}
         self.algebraics.append(z)
         self._set_transcribed(False)
         return z
 
-    def variable(self, n_rows=1, n_cols=1, grid = '',meta=None):
+    def variable(self, n_rows=1, n_cols=1, grid = '', scale=1, include_last=False, meta=None):
+        """Create a variable
+
+        Variables are unknowns in the Optimal Control problem
+        for which we seek optimal values.
+
+        Parameters
+        ----------
+        n_rows : int, optional
+            Number of rows
+        n_cols : int, optional
+            Number of columns
+        grid : string, optional
+            Default is '', resulting in a single variable available
+            over the whole optimal control horizon.
+            For MultipleShooting, 'control' can be used to
+            declare a variable that is unique to every control interval.
+            include_last determines if a unique entry is foreseen at the tf edge.
+
+
+        Returns
+        -------
+        s : :obj:`~casadi.MX`
+            A CasADi symbol representing a variable
+
+        Examples
+        --------
+
+        >>> ocp = Ocp()
+        >>> v = ocp.variable()
+        >>> x = ocp.state()
+        >>> ocp.set_der(x, v)
+        >>> ocp.set_initial(v, 3)
+        """
         # Create a placeholder symbol with a dummy name (see #25)
-        v = MX.sym("v"+str(np.random.random(1)), n_rows, n_cols)
+        L = sum([len(e) for e in self.variables.values()])
+        v = MX.sym("v"+str(L+1), n_rows, n_cols)
+        meta = merge_meta(meta, get_meta())
+        return self.register_variable(v, grid=grid, scale=scale, meta=meta, include_last=include_last)
+
+    def register_variable(self, v, grid = '', scale=1, include_last=False, meta=None):
+        if isinstance(v, list):
+            for e in v:
+                self.register_variable(e, scale=scale)
+            return
         self._meta[v] = merge_meta(meta, get_meta())
+        self._scale[v] = self._parse_scale(v, scale)
+        self._catalog[v] = {"type": 'variables', "sparsity": v.sparsity(), "grid": grid, "include_last": include_last}
+        if include_last:
+            grid+="+"
         self.variables[grid].append(v)
         self._set_transcribed(False)
         return v
+    
+    def signal_shape(self, s):
+        if s not in self._catalog:
+            raise Exception(f"{s} is not recognised as part of Ocp.")
+        data = self._catalog[s]
+        ret = {}
+        if "include_last" in data:
+            ret["include_last"] = data["include_last"]
+        if "grid" in data:
+            ret["grid"] = data["grid"]
+        return ret
 
-    def parameter(self, n_rows=1, n_cols=1, grid = '',meta=None):
-        """
-        Create a parameter
+    def parameter(self, n_rows=1, n_cols=1, grid = '', scale=1, include_last=False, meta=None):
+        """Create a parameter
+
+        Parameters are symbols of an Optimal COntrol problem
+        that are externally imposed, but not hardcoded.
+
+        The advantage of parameters over simple numbers/numerical matrices comes
+        when you need to solve multiple different Optimal Control problems.
+        Parameters avoid the need to initialize new problems form scratch all the time;
+        the problem becomes parametric.
+
+
+        Parameters
+        ----------
+        n_rows : int, optional
+            Number of rows
+        n_cols : int, optional
+            Number of columns
+        grid : string, optional
+            Default is '', resulting in a single parameter available
+            over the whole optimal control horizon. 
+            For MultipleShooting, 'control' can be used to
+            declare a parameter that is unique to every control interval.
+            include_last determines if a unique entry is foreseen at the tf edge.
+
+        Returns
+        -------
+        s : :obj:`~casadi.MX`
+            A CasADi symbol representing a parameter
+
+        Examples
+        --------
+
+        >>> ocp = Ocp()
+        >>> p = ocp.parameter()
+        >>> x = ocp.state()
+        >>> ocp.set_der(x, p)
+        >>> ocp.set_value(p, 3)
         """
         # Create a placeholder symbol with a dummy name (see #25)
-        p = MX.sym("p", n_rows, n_cols)
+        L = sum([len(e) for e in self.parameters.values()])
+        p = MX.sym("p"+str(L+1), n_rows, n_cols)
+        meta = merge_meta(meta, get_meta())
+        return self.register_parameter(p, grid=grid, scale=scale, include_last=include_last, meta=meta)
+
+    def register_parameter(self, p, grid='', scale=1, include_last=False, meta=None):
+        if isinstance(p,list):
+            for e in p:
+                self.register_parameter(e, scale=scale)
+            return
         self._meta[p] = merge_meta(meta, get_meta())
+        self._scale[p] = self._parse_scale(p, scale)
+        self._catalog[p] = {"type": 'variables', "sparsity": p.sparsity(), "grid": grid, "include_last": include_last}
+        if include_last:
+            grid+="+"
         self.parameters[grid].append(p)
         self._set_transcribed(False)
         return p
 
-    def control(self, n_rows=1, n_cols=1, order=0,meta=None):
+    def control(self, n_rows=1, n_cols=1, order=0, scale=1, meta=None):
         """Create a control signal to optimize for
 
         A control signal is parametrized as a piecewise polynomial.
@@ -282,18 +430,50 @@ class Stage:
         """
 
         if order >= 1:
-            u = self.state(n_rows, n_cols)
-            helper_u = self.control(n_rows=n_rows, n_cols=n_cols, order=order - 1)
+            u = self.state(n_rows, n_cols, scale=scale)
+            helper_u = self.control(n_rows=n_rows, n_cols=n_cols, order=order - 1, scale=scale)
             self.set_der(u, helper_u)
             return u
 
         u = MX.sym("u", n_rows, n_cols)
+        meta = merge_meta(meta, get_meta())
+        self.register_control(u, scale=scale, meta=meta)
+        return u
+
+    def register_control(self, u, scale=1, meta=None):
+        if isinstance(u,list):
+            for e in u:
+                self.register_control(e, scale=scale)
+            return
         self._meta[u] = merge_meta(meta, get_meta())
+        self._scale[u] = self._parse_scale(u, scale)
+        self._catalog[u] = {"type": 'controls', "sparsity": u.sparsity()}
         self.controls.append(u)
         self._set_transcribed(False)
         return u
 
     def set_value(self, parameter, value):
+        """Set a value for a parameter
+
+        All variables must be given a value before an optimal control problem can be solved.
+
+        Parameters
+        ----------
+        parameter : :obj:`~casadi.MX`
+            The parameter symbol to initialize
+        value : number
+            The value
+
+
+        Examples
+        --------
+
+        >>> ocp = Ocp()
+        >>> p = ocp.parameter()
+        >>> x = ocp.state()
+        >>> ocp.set_der(x, p)
+        >>> ocp.set_value(p, 3)
+        """
         if self.master is not None and self.master.is_transcribed:
             def action(parameter, value):
                 self._method.set_value(self, self.master._method, parameter, value)      
@@ -308,20 +488,57 @@ class Stage:
 
 
     def set_initial(self, var, value, priority=True):
+        """Provide an initial guess
+
+        Many Optimal Control solution methods are based on
+        iterative numerical recipes.
+        The initial guess, or starting point, may influence the
+        convergence behavior and the quality of the solution.
+
+        By default, all states, controls, and variables are initialized with zero.
+        Use set_initial to provide a non-zero initial guess.
+
+        Parameters
+        ----------
+        var : :obj:`~casadi.MX`
+            The variable, state or control symbol (shape n-by-1) to initialize
+        value : :obj:`~casadi.MX`
+            The value to initialize with. Possibilities:
+              * scalar number (repeated to fit the shape of `var` if needed)
+              * numeric matrix of shape n-by-N or n-by-(N+1) in the case of MultipleShooting
+              * CasADi symbolic expression dependent on ocp.t 
+
+        Examples
+        --------
+
+        >>> ocp = Ocp()
+        >>> x = ocp.state()
+        >>> u = ocp.control()
+        >>> ocp.set_der(x, u)
+        >>> ocp.set_initial(u, 1)
+        >>> ocp.set_initial(u, linspace(0,1,10))
+        >>> ocp.set_initial(u, sin(ocp.t))
+        """
         assert "opti" not in str(var)
         def action(var, value):
             if var not in self._meta and var not in self._placeholders:
                 raise Exception("You attempted to set the initial value of an unknown symbol: " + str(var))
             if np.any([var in p for p in self.parameters.values()]):
                 raise Exception("You attempted to set the initial value of a parameter. Did you mean ocp.set_value()? Got " + str(var))
+            if var.is_scalar():
+                # Auto-transpose
+                if isinstance(value, np.ndarray) and len(value.shape)==1:
+                    value = value.reshape((1, value.shape[0]))
+                if isinstance(value, DM) and value.shape[0]==1 and value.shape[1]>1:
+                    value = value.T
             self._initial[var] = value
             if priority:
                 self._initial.move_to_end(var, last=False)
         for_all_primitives(var, value, action, "First argument to set_initial must be a variable/signal or a simple concatenation of variables/signals")
         if self.master is not None and self.master.is_transcribed:
-            self._method.set_initial(self, self.master._method, self._initial)
+            self._method.set_initial(self._augmented, self.master._method, self._initial)
 
-    def set_der(self, state, der):
+    def set_der(self, state, der, scale=1):
         """Assign a right-hand side to a state derivative
 
         Parameters
@@ -347,6 +564,7 @@ class Stage:
             if state not in self.states and state not in self.qstates:
                 raise Exception("You used set_der on a non-state: " + str(state))
             self._state_der[state] = der
+            self._scale_der[state] = self._parse_scale(state, scale)
         for_all_primitives(state, der, action, "First argument to set_der must be a state or a simple concatenation of states")
 
     def set_next(self, state, next):
@@ -375,15 +593,23 @@ class Stage:
         for_all_primitives(state, next, action, "First argument to set_next must be a state or a simple concatenation of states")
         assert not self._state_der
 
-    def add_alg(self, constr):
+    def add_alg(self, constr, scale=1):
         self._set_transcribed(False)
-        self._alg.append(constr)
+        scale = self._parse_scale(constr, scale)
+        self._alg.append(constr/scale)
 
     def der(self, expr):
         if depends_on(expr, self.u):
             raise Exception("Dependency on controls not supported yet for stage.der")
         ode = self._ode()
-        return jtimes(expr, self.x, ode(x=self.x, u=self.u, z=self.z, p=vertcat(self.p, self.v), t=self.t)["ode"])
+        if depends_on(expr,self.t):
+            return jtimes(expr, vertcat(self.x, self.t), vertcat(ode(x=self.x, u=self.u, z=self.z, p=vertcat(self.p, self.v), t=self.t)["ode"], 1))
+        else:
+            if expr in self.states:
+                return jtimes(expr, self.x, ode.call(dict(x=self.x, u=self.u, z=self.z, p=vertcat(self.p, self.v), t=self.t),True,False)["ode"])
+            else:
+                return jtimes(expr, self.x, ode(x=self.x, u=self.u, z=self.z, p=vertcat(self.p, self.v), t=self.t)["ode"])
+
 
     def integral(self, expr, grid='inf'):
         """Compute an integral or a sum
@@ -404,7 +630,7 @@ class Stage:
         else:
             return self._create_placeholder_expr(expr, 'integral_control')
 
-    def sum(self, expr, grid='inf'):
+    def sum(self, expr, grid='control', include_last=False):
         """Compute a sum
 
         Parameters
@@ -413,11 +639,13 @@ class Stage:
             An expression to integrate over the state time domain (from t0 to tf=t0+T)
         grid : str
             Possible entries:
-                inf: the integral is performed using the integrator defined for the stage
                 control: the integral is evaluated as a sum on the control grid (start of each control interval)
                          Note that the final state is not included in this definition
         """
-        return self._create_placeholder_expr(expr, 'sum_control')
+        if include_last:
+            return self._create_placeholder_expr(expr, 'sum_control_plus')
+        else:
+            return self._create_placeholder_expr(expr, 'sum_control') 
 
     def offset(self, expr, offset):
         """Get the value of a signal at control interval current+offset
@@ -476,7 +704,7 @@ class Stage:
         self._set_transcribed(False)
         self._constraints = defaultdict(list)
 
-    def subject_to(self, constr, grid=None,include_first=True,include_last=True,meta=None):
+    def subject_to(self, constr, grid=None,include_first=True,include_last=True,scale=1,refine=1,group_refine=GroupingTechnique(),group_dim=GroupingTechnique(),group_control=GroupingTechnique(),meta=None):
         """Adds a constraint to the problem
 
         Parameters
@@ -497,7 +725,20 @@ class Stage:
                 inf: use mathematical guarantee for the whole control interval (only possible for polynomials of states and controls)
                 integrator: constrain at integrator edges
                 integrator_roots: constrain at integrator roots (e.g. collocation points excluding 0)
-
+        include_first : bool
+            Enforce constraint also at t0
+        include_last : bool or "auto"
+            Enforce constraint also at tf
+            "auto" mode will only enforce the constraint if it is not dependent on a control signal,
+            since typically control signals are not defined at tf.
+        refine : int, optional
+            Refine grid used in constraining by a certain factor with respect to the control grid
+        group_refine : GroupTechnique, optional
+            Group constraints together along the refine axis
+        group_dim : GroupTechnique, optional
+            Group vector-valued constraints along the vector dimension into a scalar constraint
+        group_control : GroupTechnique, optional
+            Group constraints together along the control grid
         Examples
         --------
 
@@ -521,7 +762,8 @@ class Stage:
             if grid != 'point': 
                 raise Exception("Expected signal expression since grid '" + grid + "' was given.")
         
-        args = {"grid": grid, "include_last": include_last, "include_first": include_first}
+        scale = self._parse_scale(constr, scale)
+        args = {"grid": grid, "include_last": include_last, "include_first": include_first, "scale": scale, "refine": refine, "group_refine": group_refine, "group_dim": group_dim, "group_control": group_control}
         self._constraints[grid].append((constr, get_meta(meta), args))
 
     def at_t0(self, expr):
@@ -593,6 +835,8 @@ class Stage:
         assert not self.is_signal(term), "An objective cannot be a signal. You must use ocp.integral or ocp.at_t0/tf to remove the time-dependence"
         self._set_transcribed(False)
         self._objective = self._objective + term
+        if not MX(term).is_scalar():
+            raise Exception("Objective terms must be scalar, got " + str(MX(term).dim())+ ".")
 
     def method(self, method):
         """Specify the transcription method
@@ -640,12 +884,13 @@ class Stage:
 
     @property
     def p(self):
-        arg = self.parameters['']+self.parameters['control']
+        arg = self.parameters['']+self.parameters['control']+self.parameters['control+']
         return MX(0, 1) if len(arg)==0 else vvcat(arg)
 
     @property
     def v(self):
-        return vvcat(self.variables['']+self.variables['control'])
+        arg = self.variables['']+self.variables['control']+self.variables['control+']
+        return MX(0, 1) if len(arg)==0 else vvcat(arg)
 
     @property
     def nx(self):
@@ -663,6 +908,34 @@ class Stage:
     def np(self):
         return self.p.numel()
 
+    @property
+    def nv(self):
+        return self.v.numel()
+
+    @property
+    def _scale_x(self):
+        return vvcat([self._scale[x] for x in self.states])
+
+    @property
+    def _scale_der_x(self):
+        return vvcat([self._scale_der[x] for x in self.states])
+
+    @property
+    def _scale_z(self):
+        return vvcat([self._scale[z] for z in self.algebraics])
+
+    @property
+    def _scale_u(self):
+        return vvcat([self._scale[u] for u in self.controls])
+
+    @property
+    def _scale_p(self):
+        return vvcat([self._scale[p] for p in self.parameters['']+self.parameters['control']+self.parameters['control+']])
+
+    @property
+    def _scale_v(self):
+        return vvcat([self._scale[v] for v in self.variables['']+self.variables['control']+self.variables['control+']])
+
     @property
     def gist(self):
         """Obtain an expression packing all information needed to obtain value/sample
@@ -685,7 +958,18 @@ class Stage:
 
         """
  
-        return depends_on(expr, vertcat(self.x, self.u, self.z, self.t, vcat(self.variables['control']+self.variables['states']), vvcat(self._inf_der.keys())))
+        return depends_on(expr, vertcat(self.x, self.u, self.z, self.t, self.DT, self.DT_control, vcat(self.parameters['control']+self.parameters['control+']), vcat(self.variables['control']+self.variables['control+']+self.variables['states']), vvcat(self._inf_der.keys())))
+
+    def is_parametric(self, expr):
+        """Does the expression depend only on parameters?
+
+        Returns
+        -------
+        res : bool
+
+        """
+ 
+        return not depends_on(expr, vertcat(self.x, self.u, self.z, self.t, vcat(self.variables['']+self.variables['control']+self.variables['control+']+self.variables['states']), vvcat(self._inf_der.keys())))
 
     def _create_placeholder_expr(self, expr, callback_name):
         """
@@ -761,9 +1045,14 @@ class Stage:
         quad = veccat(*der)
         alg = veccat(*self._alg)
         t = self.t
-        if not depends_on(vertcat(ode,alg,quad),t):
+        expr = vertcat(ode,alg,quad)
+        if not depends_on(expr,t):
             t = MX.sym('t', Sparsity(1, 1))
-        return Function('ode', [self.x, self.u, self.z, vertcat(self.p, self.v), t], [ode, alg, quad], ["x", "u", "z", "p", "t"], ["ode","alg","quad"])
+        assert not depends_on(expr,self.DT), "Your ODE right-hand-side depends on DT; not supported."
+        assert not depends_on(expr,self.DT_control), "Your ODE right-hand-side depends on DT_control; not supported."
+        ret = Function('ode', [self.x, self.u, self.z, vertcat(self.p, self.v), t], [ode, alg, quad], ["x", "u", "z", "p", "t"], ["ode","alg","quad"])
+        assert not ret.has_free()
+        return ret
 
     # Internal methods
     def _diffeq(self):
@@ -781,8 +1070,11 @@ class Stage:
             except:
                 raise Exception("ocp.set_next missing for quadrature state defined at " + str(self._meta[k]))
         quad = veccat(*val)
-        dt = MX(1,1)
-        return Function('ode', [self.x, self.u, vertcat(self.p, self.v), self.t, dt], [next, MX(), quad, MX(0, 1), MX()], ["x0", "u", "p", "t0", "DT"], ["xf","poly_coeff","qf","zf","poly_coeff_z"])
+        dt = self.DT
+        t = self.t
+        if not depends_on(vertcat(next,quad), self.t):
+            t = MX.sym('t', Sparsity(1, 1))
+        return Function('diffeq', [self.x, self.u, vertcat(self.p, self.v), t, self.DT, self.DT_control], [next, MX(), quad, MX(0, 1), MX()], ["x0", "u", "p", "t0", "DT", "DT_control"], ["xf","poly_coeff","qf","zf","poly_coeff_z"])
 
     def _expr_apply(self, expr, **kwargs):
         """
@@ -798,6 +1090,10 @@ class Stage:
     def _get_subst_set(self, **kwargs):
         subst_from = []
         subst_to = []
+        for key in ["DT","DT_control"]:
+            if key in kwargs:
+                subst_from.append(getattr(self,key))
+                subst_to.append(kwargs[key])
         if "sub" in kwargs:
             subst_from += kwargs["sub"][0]
             subst_to += kwargs["sub"][1]
@@ -830,6 +1126,10 @@ class Stage:
             p = veccat(*self.parameters['control'])
             subst_from.append(p)
             subst_to.append(kwargs["p_control"])
+        if "p_control_plus" in kwargs and self.parameters['control+']:
+            p = veccat(*self.parameters['control+'])
+            subst_from.append(p)
+            subst_to.append(kwargs["p_control_plus"])
         if "v" in kwargs and self.variables['']:
             v = veccat(*self.variables[''])
             subst_from.append(v)
@@ -838,6 +1138,10 @@ class Stage:
             v = veccat(*self.variables['control'])
             subst_from.append(v)
             subst_to.append(kwargs["v_control"])
+        if "v_control_plus" in kwargs and self.variables['control+']:
+            v = veccat(*self.variables['control+'])
+            subst_from.append(v)
+            subst_to.append(kwargs["v_control_plus"])
         if "v_states" in kwargs and self.variables['states']:
             v = veccat(*self.variables['states'])
             subst_from.append(v)
@@ -900,6 +1204,17 @@ class Stage:
         for s in self._stages:
             s._transcribe_recurse(phase=phase, **kwargs)
 
+    def _untranscribe_recurse(self, phase=1, **kwargs):
+        if self._method is not None:
+            if self is self.master:
+                self._method.main_untranscribe(self, phase=phase, **kwargs)
+            self._method.untranscribe(self, phase=phase, **kwargs)
+        else:
+            pass
+
+        for s in self._stages:
+            s._untranscribe_recurse(phase=phase, **kwargs)
+
     def _placeholders_transcribe_recurse(self, phase, placeholders):
         if self._method is not None:
             self._method.transcribe_placeholders(phase, self, placeholders)
@@ -907,6 +1222,13 @@ class Stage:
         for s in self._stages:
             s._placeholders_transcribe_recurse(phase, placeholders)
 
+    def _placeholders_untranscribe_recurse(self, phase):
+        if self._method is not None:
+            self._method.untranscribe_placeholders(phase, self)
+
+        for s in self._stages:
+            s._placeholders_untranscribe_recurse(phase)
+
     def clone(self, parent, **kwargs):
         assert self._is_original
         ret = Stage(parent, **kwargs)
@@ -966,6 +1288,8 @@ class Stage:
         ret._var_original = self._var_original
 
         ret._meta = self._meta
+        ret._scale = self._scale
+        ret._catalog = self._catalog
 
         ret._var_is_transcribed = False
         return ret
@@ -1016,7 +1340,7 @@ class Stage:
         grid : `str`
             At which points in time to sample, options are
             'control' or 'integrator' (at integrator discretization
-            level) or 'integrator_roots'.
+            level), 'integrator_roots', 'gist'.            
         refine : int, optional
             Refine grid by evaluation the polynomal of the integrater at
             intermediate points ("refine" points per interval).
@@ -1050,6 +1374,8 @@ class Stage:
                 time, res = self._grid_integrator(self, expr, grid, **kwargs)
         elif grid == 'integrator_roots':
             time, res = self._grid_integrator_roots(self, expr, grid, **kwargs)
+        elif grid == 'gist':
+            time, res = self._grid_gist(self, expr, grid, **kwargs)
         else:
             msg = "Unknown grid option: {}\n".format(grid)
             msg += "Options are: 'control' or 'integrator' with an optional extra refine=<int> argument."
@@ -1057,8 +1383,15 @@ class Stage:
 
         return placeholders(time), placeholders(res)
 
-    def _grid_control(self, stage, expr, grid, include_first=True, include_last=True, transpose=False):
+    def _grid_gist(self, stage, expr, grid, include_first=True, include_last=True, transpose=False, refine=1):
+        if hasattr(stage._method,"grid_gist"):
+            return stage._method.grid_gist(self, expr, grid, include_first=include_first, include_last=include_last, transpose=transpose, refine=refine)
+
+
+    def _grid_control(self, stage, expr, grid, include_first=True, include_last=True, transpose=False, refine=1):
         """Evaluate expression at (N + 1) control points."""
+        if hasattr(stage._method,"grid_control"):
+            return stage._method.grid_control(self, expr, grid, include_first=include_first, include_last=include_last, transpose=transpose, refine=refine)
         sub_expr = []
         ks = list(range(1, stage._method.N))
         if include_first:
@@ -1166,6 +1499,17 @@ class Stage:
         placeholders = self.master.placeholders_transcribed
         return placeholders(self._method.eval(self, expr))
 
+    @transcribed
+    def initial_value(self, expr):
+        """Get the value of an expression at initial guess
+
+        Parameters
+        ----------
+        expr : :obj:`casadi.MX`
+            Arbitrary expression containing no signals (states, controls) ...
+        """
+        return self._method.initial_value(self, expr)
+
     @transcribed
     def discrete_system(self):
         """Hack"""
